﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  [StorableClass]
  public abstract class PushExpression<T> : StatefulExpression<T> {
    [StorableConstructor]
    protected PushExpression(bool deserializing) : base(deserializing) { }

    protected PushExpression(T state, object other) : base(state) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return false;
    }

    protected void Eval(IPushStack<T> stack) {
      stack.Push(State);
    }

    public override string StringRepresentation { get { return State.ToString(); } }
  }

  [StorableClass]
  public class IntegerPushExpression : PushExpression<long> {
    [StorableConstructor]
    protected IntegerPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static IntegerPushExpression Create(IManagedPool<Expression> statefulExpressionPool, long state) {
    //  var expression = (IntegerPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public IntegerPushExpression(long state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerStack);
    }
  }

  [StorableClass]
  public class FloatPushExpression : PushExpression<double> {
    [StorableConstructor]
    protected FloatPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static FloatPushExpression Create(IManagedPool<Expression> statefulExpressionPool, double state) {
    //  var expression = (FloatPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public FloatPushExpression(double state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatStack);
    }
  }

  [StorableClass]
  public class BooleanPushExpression : PushExpression<bool> {
    [StorableConstructor]
    protected BooleanPushExpression(bool deserializing, object other) : base(deserializing) { }

    // not supported so far
    //public static BooleanPushExpression Create(IManagedPool<Expression> statefulExpressionPool, bool state) {
    //  var expression = (BooleanPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public BooleanPushExpression(bool state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanStack);
    }
  }

  [StorableClass]
  public class NamePushExpression : PushExpression<string> {
    [StorableConstructor]
    protected NamePushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static NamePushExpression Create(IManagedPool<Expression> statefulExpressionPool, string state) {
    //  var expression = (NamePushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public NamePushExpression(string state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.NameStack);
    }
  }

  [StorableClass]
  public class ExecPushExpression : PushExpression<Expression> {
    [StorableConstructor]
    protected ExecPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static ExecPushExpression Create(IManagedPool<Expression> statefulExpressionPool, Expression state) {
    //  var expression = (ExecPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public ExecPushExpression(Expression state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.ExecStack);
    }
  }

  [StorableClass]
  public class CharPushExpression : PushExpression<char> {
    [StorableConstructor]
    protected CharPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static CharPushExpression Create(IManagedPool<Expression> statefulExpressionPool, char state) {
    //  var expression = (CharPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public CharPushExpression(char state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CharStack);
    }
  }

  [StorableClass]
  public class StringPushExpression : PushExpression<string> {
    [StorableConstructor]
    protected StringPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static StringPushExpression Create(IManagedPool<Expression> statefulExpressionPool, string state) {
    //  var expression = (StringPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public StringPushExpression(string state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringStack);
    }
  }

  [StorableClass]
  public class IntegerVectorPushExpression : PushExpression<List<long>> {
    [StorableConstructor]
    protected IntegerVectorPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static IntegerVectorPushExpression Create(IManagedPool<Expression> statefulExpressionPool, List<long> state) {
    //  var expression = (IntegerVectorPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public IntegerVectorPushExpression(List<long> state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  public class FloatVectorPushExpression : PushExpression<List<double>> {
    [StorableConstructor]
    protected FloatVectorPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static FloatVectorPushExpression Create(IManagedPool<Expression> statefulExpressionPool, List<double> state) {
    //  var expression = (FloatVectorPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public FloatVectorPushExpression(List<double> state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  public class BooleanVectorPushExpression : PushExpression<List<bool>> {
    [StorableConstructor]
    protected BooleanVectorPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static BooleanVectorPushExpression Create(IManagedPool<Expression> statefulExpressionPool, List<bool> state) {
    //  var expression = (BooleanVectorPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public BooleanVectorPushExpression(List<bool> state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  public class StringVectorPushExpression : PushExpression<List<string>> {
    [StorableConstructor]
    protected StringVectorPushExpression(bool deserializing) : base(deserializing) { }

    // not supported so far
    //public static StringVectorPushExpression Create(IManagedPool<Expression> statefulExpressionPool, List<string> state) {
    //  var expression = (StringVectorPushExpression)statefulExpressionPool.Get();
    //  expression.State = state;

    //  return expression;
    //}

    public StringVectorPushExpression(List<string> state)
      : base(state, null) {
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack);
    }
  }
}