﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Takes the rest of the top item on the type stack.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorButLastExpression<T> : StatelessExpression {
    protected VectorButLastExpression() { }
    [StorableConstructor]
    protected VectorButLastExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<List<T>> vectorStack) {
      return vectorStack.IsEmpty || vectorStack.Top.Count < 2;
    }

    protected void Eval(IPushStack<List<T>> vectorStack) {
      var result = vectorStack.Top.GetRange(0, vectorStack.Top.Count - 1);
      vectorStack.SetTop(result);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].BUTLAST")]
  public class IntegerVectorButLastExpression : VectorButLastExpression<long> {
    public IntegerVectorButLastExpression() { }
    [StorableConstructor]
    protected IntegerVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].BUTLAST")]
  public class FloatVectorButLastExpression : VectorButLastExpression<double> {
    public FloatVectorButLastExpression() { }
    [StorableConstructor]
    protected FloatVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].BUTLAST")]
  public class BooleanVectorButLastExpression : VectorButLastExpression<bool> {
    public BooleanVectorButLastExpression() { }
    [StorableConstructor]
    protected BooleanVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "String[].BUTLAST")]
  public class StringVectorButLastExpression : VectorButLastExpression<string> {
    public StringVectorButLastExpression() { }
    [StorableConstructor]
    protected StringVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack);
    }
  }
}
