﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Simplifier {
  using System;
  using System.Collections.Generic;

  using Expressions;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Tree;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;

  public static class Simplifier {
    private static readonly TreeNode<Expression> NoopNode =
      new TreeNode<Expression>(ExpressionTable.GetStatelessExpression<CodeNoopExpression>());

    public static PushProgram Simplify(PushProgram program, IReadOnlyPushConfiguration config, Func<PushProgram, double> evaluator) {
      var bestResult = evaluator(program);
      //var noopFreeProgram = RemoveNoops(program, config);
      var tree = program.ToTree();
      var optimizedTree = RemoveUnnecessaryExpressions(tree, tree, ref bestResult, evaluator);

      var optimizedPrgram = optimizedTree.ToPushProgram(e => !ReferenceEquals(e, NoopNode.Value));
      return optimizedPrgram; // RemoveNoops(optimizedPrgram, config);
    }

    // not working so far, an expression can only be a noop if it is a noop for all examples of a problem
    private static PushProgram RemoveNoops(PushProgram pushProgram, IReadOnlyPushConfiguration config) {
      var expressions = new List<Expression>();
      var interpreter = new PushInterpreter(config);

      interpreter.Run(pushProgram, true);

      while (interpreter.CanStep) {
        var isNoop = interpreter.ExecStack.Top.IsNoop(interpreter);
        if (!isNoop)
          expressions.Add(interpreter.ExecStack.Top);

        interpreter.Step();
      }

      return new PushProgram(expressions);
    }

    private static TreeNode<Expression> RemoveUnnecessaryExpressions(Tree<Expression> program, TreeNode<Expression> node, ref double bestResult, Func<PushProgram, double> evaluator) {
      for (var i = 0; i < node.Children.Count; i++) {
        var child = node.Children[i];
        node.ReplaceNode(i, NoopNode);

        var currentProgram = program.ToPushProgram();
        var currentResult = evaluator(currentProgram);

        if (currentResult <= bestResult) {
          bestResult = currentResult;
        } else if (child.Value.IsProgram) {
          node.ReplaceNode(i, child);
          var subProgram = RemoveUnnecessaryExpressions(program, child, ref bestResult, evaluator);
          node.ReplaceNode(i, subProgram);
        } else {
          node.ReplaceNode(i, child);
        }
      }

      return node;
    }
  }
}
