﻿using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Stack {
  using HeuristicLab.Problems.ProgramSynthesis.Push.Constants;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;

  public class InterpreterStackStringifier {

    public static IEnumerable<string> StringifyStack(IPushInterpreter interpreter, StackTypes type) {
      switch (type) {
        case StackTypes.Boolean: return StringifyLiteralStack(interpreter.BooleanStack);
        case StackTypes.Integer: return StringifyLiteralStack(interpreter.IntegerStack);
        case StackTypes.Float: return StringifyLiteralStack(interpreter.FloatStack);
        case StackTypes.Char: return StringifyLiteralStack(interpreter.CharStack);
        case StackTypes.String: return StringifyLiteralStack(interpreter.StringStack, "\"", "\"");
        case StackTypes.Name: return StringifyLiteralStack(interpreter.NameStack);
        case StackTypes.Code: return StringifyLiteralStack(interpreter.CodeStack);
        case StackTypes.Exec: return StringifyLiteralStack(interpreter.ExecStack);
        case StackTypes.Print: return StringifyLiteralStack(interpreter.PrintStack);

        case StackTypes.BooleanVector: return StringifyVectorStack(interpreter.BooleanVectorStack);
        case StackTypes.IntegerVector: return StringifyVectorStack(interpreter.IntegerVectorStack);
        case StackTypes.FloatVector: return StringifyVectorStack(interpreter.FloatVectorStack);
        case StackTypes.StringVector: return StringifyVectorStack(interpreter.StringVectorStack, "\"", "\"");

        default: return Enumerable.Empty<string>();
      }
    }

    private static IEnumerable<string> StringifyLiteralStack<T>(IPushStack<T> stack, string prefix = "", string postfix = "") {
      return stack.Reverse().Select(x => prefix + x.ToString().Replace(PushEnvironment.NewLine, "\\n") + postfix);
    }

    private static IEnumerable<string> StringifyVectorStack<T>(IPushStack<List<T>> stack, string prefix = "", string postfix = "") {
      var separator = PushEnvironment.VectorSeparatorSymbol.ToString();
      return stack
        .Reverse()
        .Select(x => PushEnvironment.VectorStartSymbol + string.Join(separator, x.Select(_ => prefix + _.ToString() + postfix)) + PushEnvironment.VectorEndSymbol);
    }
  }
}
